import os
import sys
import time
from contextlib import suppress

from blessed import Terminal

from gpustat import __version__
from gpustat.core import GPUStatCollection


SHTAB_PREAMBLE = {
    'zsh': '''\
# % gpustat -i <TAB>
# float
# % gpustat -i -<TAB>
# option
# -a                   Display all gpu properties above
# ...
_complete_for_one_or_zero() {
  if [[ ${words[CURRENT]} == -* ]]; then
    # override the original options
    _shtab_gpustat_options=(${words[CURRENT - 1]} $_shtab_gpustat_options)
    _arguments -C $_shtab_gpustat_options
  else
    eval "${@[-1]}"
  fi
}
'''
}


def zsh_choices_to_complete(choices, tag='', description=''):
    '''Change choices to complete for zsh.

    https://github.com/zsh-users/zsh/blob/master/Etc/completion-style-guide#L224
    '''
    complete = 'compadd - ' + ' '.join(filter(len, choices))
    if description == '':
        description = tag
    if tag != '':
        complete = '_wanted ' + tag + ' expl ' + description + ' ' + complete
    return complete


def get_complete_for_one_or_zero(input):
    '''Get shell complete for nargs='?'. Now only support zsh.'''
    output = {}
    for sh, complete in input.items():
        if sh == 'zsh':
            output[sh] = "_complete_for_one_or_zero '" + complete + "'"
    return output


def print_gpustat(*, id=None, json=False, debug=False, **kwargs):
    '''Display the GPU query results into standard output.'''
    try:
        gpu_stats = GPUStatCollection.new_query(debug=debug, id=id)
    except Exception as e:
        sys.stderr.write('Error on querying NVIDIA devices. '
                         'Use --debug flag to see more details.\n')
        term = Terminal(stream=sys.stderr)
        sys.stderr.write(term.red(str(e)) + '\n')

        if debug:
            sys.stderr.write('\n')
            try:
                import traceback
                traceback.print_exc(file=sys.stderr)
            except Exception:
                # NVMLError can't be processed by traceback:
                #   https://bugs.python.org/issue28603
                # as a workaround, simply re-throw the exception
                raise e

        sys.stderr.flush()
        sys.exit(1)

    if json:
        gpu_stats.print_json(sys.stdout)
    else:
        gpu_stats.print_formatted(sys.stdout, **kwargs)


def loop_gpustat(interval=1.0, **kwargs):
    term = Terminal()

    with term.fullscreen():
        while 1:
            try:
                query_start = time.time()

                # Move cursor to (0, 0) but do not restore original cursor loc
                print(term.move(0, 0), end='')
                print_gpustat(eol_char=term.clear_eol + os.linesep, **kwargs)
                print(term.clear_eos, end='')

                query_duration = time.time() - query_start
                sleep_duration = interval - query_duration
                if sleep_duration > 0:
                    time.sleep(sleep_duration)
            except KeyboardInterrupt:
                return 0


def main(*argv):
    if not argv:
        argv = list(sys.argv)

    # attach SIGPIPE handler to properly handle broken pipe
    try:  # sigpipe not available under windows. just ignore in this case
        import signal
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    except Exception as e:
        pass
    # arguments to gpustat
    import argparse
    try:
        import shtab
    except ImportError:
        from . import _shtab as shtab
    parser = argparse.ArgumentParser('gpustat')
    shtab.add_argument_to(parser, preamble=SHTAB_PREAMBLE)

    def nonnegative_int(value):
        value = int(value)
        if value < 0:
            raise argparse.ArgumentTypeError(
                "Only non-negative integers are allowed.")
        return value

    parser_color = parser.add_mutually_exclusive_group()
    parser_color.add_argument('--force-color', '--color', action='store_true',
                              help='Force to output with colors')
    parser_color.add_argument('--no-color', action='store_true',
                              help='Suppress colored output')
    parser.add_argument('--id', help='Target a specific GPU (index).')
    parser.add_argument('-a', '--show-all', action='store_true',
                        help='Display all gpu properties above')
    parser.add_argument('-c', '--show-cmd', action='store_true',
                        help='Display cmd name of running process')
    parser.add_argument(
        '-f', '--show-full-cmd', action='store_true', default=False,
        help='Display full command and cpu stats of running process'
    )
    parser.add_argument('-u', '--show-user', action='store_true',
                        help='Display username of running process')
    parser.add_argument('-p', '--show-pid', action='store_true',
                        help='Display PID of running process')
    parser.add_argument('-F', '--show-fan-speed', '--show-fan',
                        action='store_true', help='Display GPU fan speed')
    codec_choices = ['', 'enc', 'dec', 'enc,dec']
    parser.add_argument(
        '-e', '--show-codec', nargs='?', const='enc,dec', default='',
        choices=codec_choices,
        help='Show encoder/decoder utilization'
    ).complete = get_complete_for_one_or_zero(  # type: ignore
        {'zsh': zsh_choices_to_complete(codec_choices, 'codec')}
    )
    power_choices = ['', 'draw', 'limit', 'draw,limit', 'limit,draw']
    parser.add_argument(
        '-P', '--show-power', nargs='?', const='draw,limit',
        choices=power_choices,
        help='Show GPU power usage or draw (and/or limit)'
    ).complete = get_complete_for_one_or_zero(  # type: ignore
        {'zsh': zsh_choices_to_complete(power_choices, 'power')}
    )
    parser.add_argument('--json', action='store_true', default=False,
                        help='Print all the information in JSON format')
    parser.add_argument(
        '-i', '--interval', '--watch', nargs='?', type=float, default=0,
        help='Use watch mode if given; seconds to wait between updates'
    ).complete = get_complete_for_one_or_zero({'zsh': '_numbers float'})  # type: ignore
    parser.add_argument(
        '--no-header', dest='show_header', action='store_false', default=True,
        help='Suppress header message'
    )
    parser.add_argument(
        '--gpuname-width', type=nonnegative_int, default=None,
        help='The width at which GPU names will be displayed.'
    )
    parser.add_argument(
        '--debug', action='store_true', default=False,
        help='Allow to print additional informations for debugging.'
    )
    parser.add_argument(
        '--no-processes', dest='no_processes', action='store_true',
        help='Do not display running process information (memory, user, etc.)'
    )
    parser.add_argument('-v', '--version', action='version',
                        version=('gpustat %s' % __version__))
    args = parser.parse_args(argv[1:])
    # TypeError: GPUStatCollection.print_formatted() got an unexpected keyword argument 'print_completion'
    with suppress(AttributeError):
        del args.print_completion  # type: ignore
    if args.show_all:
        args.show_cmd = True
        args.show_user = True
        args.show_pid = True
        args.show_fan_speed = True
        args.show_codec = 'enc,dec'
        args.show_power = 'draw,limit'
    del args.show_all

    if args.interval is None:  # with default value
        args.interval = 1.0
    if args.interval > 0:
        args.interval = max(0.1, args.interval)
        if args.json:
            sys.stderr.write("Error: --json and --interval/-i "
                             "can't be used together.\n")
            sys.exit(1)

        loop_gpustat(**vars(args))
    else:
        del args.interval
        print_gpustat(**vars(args))


if __name__ == '__main__':
    main(*sys.argv)



"""Imports pynvml with sanity checks and custom patches."""

import warnings
import functools
import os
import sys
import textwrap

# If this environment variable is set, we will bypass pynvml version validation
# so that legacy pynvml (nvidia-ml-py3) can be used. This would be useful
# in a case where there are conflicts on pynvml dependencies.
# However, beware that pynvml might produce wrong results (see #107).
ALLOW_LEGACY_PYNVML = os.getenv("ALLOW_LEGACY_PYNVML", "")
ALLOW_LEGACY_PYNVML = ALLOW_LEGACY_PYNVML.lower() not in ('false', '0', '')


try:
    # Check pynvml version: we require 11.450.129 or newer.
    # https://github.com/wookayin/gpustat/pull/107
    import pynvml
    if not (
        # Requires nvidia-ml-py >= 11.460.79
        hasattr(pynvml, 'NVML_BRAND_NVIDIA_RTX') or
        # Requires nvidia-ml-py >= 11.450.129, < 11.510.69
        hasattr(pynvml, 'nvmlDeviceGetComputeRunningProcesses_v2')
    ) and not ALLOW_LEGACY_PYNVML:
        raise ImportError("pynvml library is outdated.")

    if not hasattr(pynvml, '_nvmlGetFunctionPointer'):
        # Unofficial pynvml from @gpuopenanalytics/pynvml, see #153
        import pynvml.nvml as pynvml

except (ImportError, SyntaxError, RuntimeError) as e:
    _pynvml = sys.modules.get('pynvml', None)

    raise ImportError(textwrap.dedent(
        """\
        pynvml is missing or an outdated version is installed.

        We require nvidia-ml-py>=11.450.129, and the official NVIDIA python bindings
        should be used; neither nvidia-ml-py3 nor gpuopenanalytics/pynvml.
        For more details, please refer to: https://github.com/wookayin/gpustat/issues/107

        The root cause: """ + str(e) +
        """

        Your pynvml installation: """ + repr(_pynvml) +
        """

        -----------------------------------------------------------
        Please reinstall `gpustat`:

        $ pip install --force-reinstall gpustat

        If it still does not fix the problem, please uninstall pynvml packages and reinstall nvidia-ml-py manually:

        $ pip uninstall nvidia-ml-py3 pynvml
        $ pip install --force-reinstall --ignore-installed 'nvidia-ml-py'
        """)) from e


# Monkey-patch nvml due to breaking changes in pynvml.
# See #107, #141, and test_gpustat.py for more details.

_original_nvmlGetFunctionPointer = pynvml._nvmlGetFunctionPointer
_original_nvmlDeviceGetMemoryInfo = pynvml.nvmlDeviceGetMemoryInfo


class pynvml_monkeypatch:

    @staticmethod  # Note: must be defined as a staticmethod to allow mocking.
    def original_nvmlGetFunctionPointer(name):
        return _original_nvmlGetFunctionPointer(name)

    FUNCTION_FALLBACKS = {
        # for pynvml._nvmlGetFunctionPointer
        'nvmlDeviceGetComputeRunningProcesses_v3': 'nvmlDeviceGetComputeRunningProcesses_v2',
        'nvmlDeviceGetGraphicsRunningProcesses_v3': 'nvmlDeviceGetGraphicsRunningProcesses_v2',
    }

    @staticmethod
    @functools.wraps(pynvml._nvmlGetFunctionPointer)
    def _nvmlGetFunctionPointer(name):
        """Our monkey-patched pynvml._nvmlGetFunctionPointer().

        See also:
            test_gpustat::NvidiaDriverMock for test scenarios. See #107.
        """
        M = pynvml_monkeypatch

        try:
            ret = M.original_nvmlGetFunctionPointer(name)
            return ret
        except pynvml.NVMLError_FunctionNotFound:  # type: ignore
            if name in M.FUNCTION_FALLBACKS:
                # Lack of ...Processes_v3 APIs happens for
                # OLD drivers < 510.39.01 && pynvml >= 11.510, where
                # we fallback to v2 APIs. (see #107 for more details)

                ret = M.original_nvmlGetFunctionPointer(
                    M.FUNCTION_FALLBACKS[name]
                )
                # populate the cache, so this handler won't get executed again
                pynvml._nvmlGetFunctionPointer_cache[name] = ret

            else:
                # Unknown case, cannot handle. re-raise again
                raise

        return ret

    @staticmethod  # Note: must be defined as a staticmethod to allow mocking.
    def original_nvmlDeviceGetMemoryInfo(*args, **kwargs):
        return _original_nvmlDeviceGetMemoryInfo(*args, **kwargs)

    has_memoryinfo_v2 = None

    @staticmethod
    @functools.wraps(pynvml.nvmlDeviceGetMemoryInfo)
    def nvmlDeviceGetMemoryInfo(handle):
        """A patched version of nvmlDeviceGetMemoryInfo.

        This tries `version=N.nvmlMemory_v2` if the nvmlDeviceGetMemoryInfo_v2
        function is available (for driver >= 515), or fallback to the legacy
        v1 API for (driver < 515) to yield a correct result. See #141.
        """
        M = pynvml_monkeypatch

        if M.has_memoryinfo_v2 is None:
            try:
                pynvml._nvmlGetFunctionPointer("nvmlDeviceGetMemoryInfo_v2")
                M.has_memoryinfo_v2 = True
            except pynvml.NVMLError_FunctionNotFound:  # type: ignore
                M.has_memoryinfo_v2 = False

        if hasattr(pynvml, 'nvmlMemory_v2'):  # pynvml >= 11.510.69
            try:
                memory = M.original_nvmlDeviceGetMemoryInfo(
                    handle, version=pynvml.nvmlMemory_v2)
            except pynvml.NVMLError_FunctionNotFound:  # type: ignore
                # pynvml >= 11.510 but driver is old (<515.39)
                memory = M.original_nvmlDeviceGetMemoryInfo(handle)
        else:
            if M.has_memoryinfo_v2:
                warnings.warn(
                    "Your NVIDIA driver requires a compatible version of "
                    "pynvml (>= 11.510.69) installed to display the correct "
                    "memory usage information (See #141 for more details). "
                    "Please try `pip install --upgrade nvidia-ml-py`.",
                    category=UserWarning)
            memory = M.original_nvmlDeviceGetMemoryInfo(handle)

        return memory


setattr(pynvml, '_nvmlGetFunctionPointer', pynvml_monkeypatch._nvmlGetFunctionPointer)
setattr(pynvml, 'nvmlDeviceGetMemoryInfo', pynvml_monkeypatch.nvmlDeviceGetMemoryInfo)


__all__ = ['pynvml']



"""
The gpustat module.
"""

# isort: skip_file
try:
    from ._version import version as __version__
    from ._version import version_tuple as __version_tuple__
except (ImportError, AttributeError) as ex:
    raise ImportError(
        "Unable to find `gpustat.__version__` string. "
        "Please try reinstalling gpustat; or if you are on a development "
        "version, then run `pip install -e .` and try again."
    ) from ex

from .core import GPUStat, GPUStatCollection
from .core import new_query, gpu_count, is_available
from .cli import print_gpustat, main


__all__ = (
    '__version__',
    'GPUStat',
    'GPUStatCollection',
    'new_query',
    'gpu_count',
    'is_available',
    'print_gpustat',
    'main',
)



"""
Unit or integration tests for gpustat
"""
# flake8: ignore=E501

import ctypes
import os
import shlex
import sys
import types
from collections import namedtuple
from io import StringIO
from typing import Any

import psutil
import pytest
from mockito import ANY, mock, unstub, when, when2

import gpustat
from gpustat.nvml import pynvml, pynvml_monkeypatch

MB = 1024 * 1024


def remove_ansi_codes(s):
    import re
    s = re.compile(r'\x1b[^m]*m').sub('', s)
    s = re.compile(r'\x0f').sub('', s)
    return s

# -----------------------------------------------------------------------------

mock_gpu_handles = [types.SimpleNamespace(value='mock-handle-%d' % i, index=i)
                    for i in range(3)]

def _configure_mock(N=pynvml,
                    _scenario_nonexistent_pid=False,  # GH-95
                    _scenario_failing_one_gpu=None,   # GH-125, GH-81
                    ):
    """Define mock behaviour for pynvml and psutil.{Process,virtual_memory}."""

    # without following patch, unhashable NVMLError makes unit test crash
    N.NVMLError.__hash__ = lambda _: 0
    assert issubclass(N.NVMLError, BaseException)

    unstub(N)  # reset all the stubs

    when(N).nvmlInit().thenReturn()
    when(N).nvmlShutdown().thenReturn()
    when(N).nvmlSystemGetDriverVersion().thenReturn('415.27.mock')

    when(N)._nvmlGetFunctionPointer(...).thenCallOriginalImplementation()

    NUM_GPUS = 3
    when(N).nvmlDeviceGetCount().thenReturn(NUM_GPUS)

    def _return_or_raise(v):
        """Return a callable for thenAnswer() to let exceptions re-raised."""
        def _callable(*args, **kwargs):
            del args, kwargs
            if isinstance(v, Exception):
                raise v
            return v
        return _callable

    for i in range(NUM_GPUS):
        handle = mock_gpu_handles[i]
        if _scenario_failing_one_gpu and i == 2:  # see #81, #125
            assert (_scenario_failing_one_gpu is N.NVMLError_Unknown or
                    _scenario_failing_one_gpu is N.NVMLError_GpuIsLost)
            handle = _scenario_failing_one_gpu()  # see 81

        when(N).nvmlDeviceGetHandleByIndex(i)\
            .thenAnswer(_return_or_raise(handle))
        when(N).nvmlDeviceGetIndex(handle)\
            .thenReturn(i)
        when(N).nvmlDeviceGetName(handle)\
            .thenReturn({
                0: 'GeForce GTX TITAN 0',
                1: 'GeForce GTX TITAN 1',
                2: 'GeForce RTX 2',
            }[i].encode())
        when(N).nvmlDeviceGetUUID(handle)\
            .thenReturn({
                0: b'GPU-10fb0fbd-2696-43f3-467f-d280d906a107',
                1: b'GPU-d1df4664-bb44-189c-7ad0-ab86c8cb30e2',
                2: b'GPU-50205d95-57b6-f541-2bcb-86c09afed564',
            }[i])

        when(N).nvmlDeviceGetTemperature(handle, N.NVML_TEMPERATURE_GPU)\
            .thenReturn([80, 36, 71][i])
        when(N).nvmlDeviceGetFanSpeed(handle)\
            .thenReturn([16, 53, 100][i])
        when(N).nvmlDeviceGetPowerUsage(handle)\
            .thenAnswer(_return_or_raise({
                0: 125000, 1: N.NVMLError_NotSupported(), 2: 250000
            }[i]))
        when(N).nvmlDeviceGetEnforcedPowerLimit(handle)\
            .thenAnswer(_return_or_raise({
                0: 250000, 1: 250000, 2: N.NVMLError_NotSupported()
            }[i]))

        # see also: NvidiaDriverMock
        mock_memory_t = namedtuple("Memory_t", ['total', 'used'])  # c_nvmlMemory_t
        when(N).nvmlDeviceGetMemoryInfo(handle)\
            .thenAnswer(_return_or_raise({
                0: mock_memory_t(total=12883853312, used=8000*MB),
                1: mock_memory_t(total=12781551616, used=9000*MB),
                2: mock_memory_t(total=12781551616, used=0),
            }[i]))
        # this mock function assumes <510.39 behavior (#141)
        when(N, strict=False)\
            .nvmlDeviceGetMemoryInfo(handle, version=ANY())\
            .thenRaise(N.NVMLError_FunctionNotFound)

        mock_utilization_t = namedtuple("Utilization_t", ['gpu', 'memory'])
        when(N).nvmlDeviceGetUtilizationRates(handle)\
            .thenAnswer(_return_or_raise({
                0: mock_utilization_t(gpu=76, memory=0),
                1: mock_utilization_t(gpu=0, memory=0),
                2: N.NVMLError_NotSupported(),  # Not Supported
            }[i]))

        when(N).nvmlDeviceGetEncoderUtilization(handle)\
            .thenAnswer(_return_or_raise({
                0: [88, 167000],  # [value, sample_rate]
                1: [0, 167000],   # [value, sample_rate]
                2: N.NVMLError_NotSupported(),  # Not Supported
            }[i]))
        when(N).nvmlDeviceGetDecoderUtilization(handle)\
            .thenAnswer(_return_or_raise({
                0: [67, 167000],  # [value, sample_rate]
                1: [0, 167000],   # [value, sample_rate]
                2: N.NVMLError_NotSupported(),  # Not Supported
            }[i]))

        # running process information: a bit annoying...
        mock_process_t = namedtuple("Process_t", ['pid', 'usedGpuMemory'])

        if _scenario_nonexistent_pid:
            mock_processes_gpu2_erratic = [
                mock_process_t(99999, 9999*MB),
                mock_process_t(99995, 9995*MB),   # see issue #95
            ]
        else:
            mock_processes_gpu2_erratic = N.NVMLError_NotSupported()

        # see NvidiaDriverMock as well
        when(N).nvmlDeviceGetComputeRunningProcesses(handle)\
            .thenAnswer(_return_or_raise({
                0: [mock_process_t(48448, 4000*MB), mock_process_t(153223, 4000*MB)],
                1: [mock_process_t(192453, 3000*MB), mock_process_t(194826, 6000*MB)],
                2: mock_processes_gpu2_erratic,   # Not Supported or non-existent
            }[i]))

        when(N).nvmlDeviceGetGraphicsRunningProcesses(handle)\
            .thenAnswer(_return_or_raise({
                0: [mock_process_t(48448, 4000*MB)],
                1: [],
                2: N.NVMLError_NotSupported(),
            }[i]))

    # for psutil
    mock_pid_map = {   # mock/stub information for psutil...
        48448:  ('user1', 'python', 85.25, 3.1415),
        154213: ('user1', 'caffe', 16.89, 100.00),
        38310:  ('user3', 'python', 26.23, 99.9653),
        153223: ('user2', 'python', 15.25, 0.0000),
        194826: ('user3', 'caffe', 0.0, 12.5236),
        192453: ('user1', 'torch', 123.2, 0.7312),
    }
    assert 99999 not in mock_pid_map, 'scenario_nonexistent_pid'
    assert 99995 not in mock_pid_map, 'scenario_nonexistent_pid (#95)'

    def _MockedProcess(pid):
        if pid not in mock_pid_map:
            if pid == 99995:
                # simulate a bug reported in #95
                raise FileNotFoundError("/proc/99995/stat")
            else:
                # for a process that does not exist, NoSuchProcess is the
                # type of exceptions supposed to be raised by psutil
                raise psutil.NoSuchProcess(pid=pid)
        username, cmdline, cpuutil, memutil = mock_pid_map[pid]
        p: Any = mock(strict=True)   # psutil.Process
        p.username = lambda: username
        p.cmdline = lambda: [cmdline]
        p.cpu_percent = lambda: cpuutil
        p.memory_percent = lambda: memutil
        p.pid = pid
        return p

    when(psutil).Process(...)\
        .thenAnswer(_MockedProcess)
    when(psutil).virtual_memory()\
        .thenReturn(mock_memory_t(total=8589934592, used=0))


MOCK_EXPECTED_OUTPUT_DEFAULT = os.linesep.join("""\
[0] GeForce GTX TITAN 0 | 80°C,  76 % |  8000 / 12287 MB | user1(4000M) user2(4000M)
[1] GeForce GTX TITAN 1 | 36°C,   0 % |  9000 / 12189 MB | user1(3000M) user3(6000M)
[2] GeForce RTX 2       | 71°C,  ?? % |     0 / 12189 MB | (Not Supported)
""".splitlines())  # noqa: E501

MOCK_EXPECTED_OUTPUT_FULL = os.linesep.join("""\
[0] GeForce GTX TITAN 0 | 80°C,  16 %,  76 % (E:  88 %  D:  67 %),  125 / 250 W |  8000 / 12287 MB | user1:python/48448(4000M) user2:python/153223(4000M)
[1] GeForce GTX TITAN 1 | 36°C,  53 %,   0 % (E:   0 %  D:   0 %),   ?? / 250 W |  9000 / 12189 MB | user1:torch/192453(3000M) user3:caffe/194826(6000M)
[2] GeForce RTX 2       | 71°C, 100 %,  ?? % (E:  ?? %  D:  ?? %),  250 /  ?? W |     0 / 12189 MB | (Not Supported)
""".splitlines())  # noqa: E501

MOCK_EXPECTED_OUTPUT_FULL_PROCESS = os.linesep.join("""\
[0] GeForce GTX TITAN 0 | 80°C,  16 %,  76 % (E:  88 %  D:  67 %),  125 / 250 W |  8000 / 12287 MB | user1:python/48448(4000M) user2:python/153223(4000M)
 ├─  48448 (  85%,  257MB): python
 └─ 153223 (  15%,     0B): python
[1] GeForce GTX TITAN 1 | 36°C,  53 %,   0 % (E:   0 %  D:   0 %),   ?? / 250 W |  9000 / 12189 MB | user1:torch/192453(3000M) user3:caffe/194826(6000M)
 ├─ 192453 ( 123%,   59MB): torch
 └─ 194826 (   0%, 1025MB): caffe
[2] GeForce RTX 2       | 71°C, 100 %,  ?? % (E:  ?? %  D:  ?? %),  250 /  ?? W |     0 / 12189 MB | (Not Supported)
""".splitlines())  # noqa: E501

MOCK_EXPECTED_OUTPUT_NO_PROCESSES = os.linesep.join("""\
[0] GeForce GTX TITAN 0 | 80°C,  76 % |  8000 / 12287 MB
[1] GeForce GTX TITAN 1 | 36°C,   0 % |  9000 / 12189 MB
[2] GeForce RTX 2       | 71°C,  ?? % |     0 / 12189 MB
""".splitlines())  # noqa: E501

# -----------------------------------------------------------------------------


@pytest.fixture
def scenario_basic():
    _configure_mock()


@pytest.fixture
def scenario_nonexistent_pid():
    _configure_mock(_scenario_nonexistent_pid=True)


@pytest.fixture
def scenario_failing_one_gpu(request: pytest.FixtureRequest):
    # request.param should be either NVMLError_Unknown or NVMLError_GpuIsLost
    _configure_mock(_scenario_failing_one_gpu=request.param)
    return dict(expected_message={
        pynvml.NVMLError_GpuIsLost: 'GPU is lost',
        pynvml.NVMLError_Unknown: 'Unknown Error',
    }[request.param])


@pytest.fixture
def nvidia_driver_version(request: pytest.FixtureRequest):
    """See NvidiaDriverMock."""

    nvidia_mock: NvidiaDriverMock = request.param
    nvidia_mock(pynvml)

    if nvidia_mock.name.startswith('430'):
        # AssertionError: gpustat will print (Not Supported) in this case
        request.node.add_marker(pytest.mark.xfail(
            reason="nvmlDeviceGetComputeRunningProcesses_v2 does not exist"))

    yield nvidia_mock


class NvidiaDriverMock:
    """Simulate the behavior of nvml's low-level functions according to a
    specific nvidia driver versions, with backward compatibility in concern.
    In all the scenarios, gpustat should work well with a compatible version
    of pynvml installed.

    For what has changed on the nvidia driver side (a non-exhaustive list), see
    https://github.com/NVIDIA/nvidia-settings/blame/main/src/nvml.h
    https://github.com/NVIDIA/nvidia-settings/blame/main/src/libXNVCtrlAttributes/NvCtrlAttributesPrivate.h

    Noteworthy changes of nvml driviers:
        450.66:    nvmlDeviceGetComputeRunningProcesses_v2
        510.39.01: nvmlDeviceGetComputeRunningProcesses_v3  (_v2 removed)
                   nvmlDeviceGetMemoryInfo_v2

    Relevant github issues:
        #107: nvmlDeviceGetComputeRunningProcesses_v2 added
        #141: nvmlDeviceGetMemoryInfo (v1) broken for 510.39.01+
    """
    INSTANCES = []

    def __init__(self, name, **kwargs):
        self.name = name
        self.feat = kwargs

    def __call__(self, N):
        self.mock_processes(N)
        self.mock_memoryinfo(N)

    def mock_processes(self, N):
        when(N).nvmlDeviceGetComputeRunningProcesses(...).thenCallOriginalImplementation()
        when(N).nvmlDeviceGetGraphicsRunningProcesses(...).thenCallOriginalImplementation()
        when(N).nvmlSystemGetDriverVersion().thenReturn(self.name)

        def process_t(pid, usedGpuMemory):
            return pynvml.c_nvmlProcessInfo_t(
                pid=ctypes.c_uint(pid),
                usedGpuMemory=ctypes.c_ulonglong(usedGpuMemory),
            )

        # more low-level mocking for
        # nvmlDeviceGetComputeRunningProcesses_{v2, v3} & c_nvmlProcessInfo_t
        def _nvmlDeviceGetComputeRunningProcesses_v2(handle, c_count, c_procs):
            # handle: SimpleNamespace (see _configure_mock)
            if c_count._obj.value == 0:
                return pynvml.NVML_ERROR_INSUFFICIENT_SIZE
            else:
                c_count._obj.value = 2
                if handle.index == 0:
                    c = process_t(pid=48448, usedGpuMemory=4000*MB); c_procs[0] = c
                    c = process_t(pid=153223, usedGpuMemory=4000*MB); c_procs[1] = c
                elif handle.index == 1:
                    c = process_t(pid=192453, usedGpuMemory=3000*MB); c_procs[0] = c
                    c = process_t(pid=194826, usedGpuMemory=6000*MB); c_procs[1] = c
                else:
                    return pynvml.NVML_ERROR_NOT_SUPPORTED
            return pynvml.NVML_SUCCESS

        def _nvmlDeviceGetGraphicsRunningProcesses_v2(handle, c_count, c_procs):
            if c_count._obj.value == 0:
                return pynvml.NVML_ERROR_INSUFFICIENT_SIZE
            else:
                if handle.index == 0:
                    c_count._obj.value = 1
                    c = process_t(pid=48448, usedGpuMemory=4000*MB); c_procs[0] = c
                elif handle.index == 1:
                    c_count._obj.value = 0
                else:
                    return pynvml.NVML_ERROR_NOT_SUPPORTED
            return pynvml.NVML_SUCCESS

        # Note: N._nvmlGetFunctionPointer might have been monkey-patched,
        # so this mock should decorate the underlying, unwrapped raw function,
        # NOT a monkey-patched version of pynvml._nvmlGetFunctionPointer.
        for v in [1, 2, 3]:
            _v = f'_v{v}' if v != 1 else ''   # backward compatible v3 -> v2
            stub = when2(pynvml_monkeypatch.original_nvmlGetFunctionPointer,
                         f'nvmlDeviceGetComputeRunningProcesses{_v}')
            if v <= self.nvmlDeviceGetComputeRunningProcesses_v:
                stub.thenReturn(_nvmlDeviceGetComputeRunningProcesses_v2)
            else:
                stub.thenRaise(pynvml.NVMLError(pynvml.NVML_ERROR_FUNCTION_NOT_FOUND))

            stub = when2(pynvml_monkeypatch.original_nvmlGetFunctionPointer,
                         f'nvmlDeviceGetGraphicsRunningProcesses{_v}')
            if v <= self.nvmlDeviceGetComputeRunningProcesses_v:
                stub.thenReturn(_nvmlDeviceGetGraphicsRunningProcesses_v2)
            else:
                stub.thenRaise(pynvml.NVMLError(pynvml.NVML_ERROR_FUNCTION_NOT_FOUND))

    def mock_memoryinfo(self, N):
        nvmlMemory_v2 = 0x02000028
        if self.nvmlDeviceGetMemoryInfo_v == 1:
            mock_memory_t = namedtuple(
                "c_nvmlMemory_t",
                ['total', 'used'],
            )
        elif self.nvmlDeviceGetMemoryInfo_v == 2:
            mock_memory_t = namedtuple(
                "c_nvmlMemory_v2_t",
                ['version', 'total', 'reserved', 'free', 'used'],
            )
            mock_memory_t.__new__.__defaults__ = (nvmlMemory_v2, 0, 0, 0, 0)
        else:
            raise NotImplementedError

        # simulates drivers >= 510.39, where memoryinfo v2 is introduced
        if self.nvmlDeviceGetMemoryInfo_v == 2:
            for handle in mock_gpu_handles:
                # a correct API requires version=... parameter
                # this assumes nvidia driver is also recent enough.
                when(pynvml_monkeypatch, strict=False)\
                    .original_nvmlDeviceGetMemoryInfo(handle, version=nvmlMemory_v2)\
                    .thenReturn({
                        0: mock_memory_t(total=12883853312, used=8000*MB),
                        1: mock_memory_t(total=12781551616, used=9000*MB),
                        2: mock_memory_t(total=12781551616, used=0),
                    }[handle.index])
                # simulate #141: without the v2 parameter, gives wrong result
                when(pynvml_monkeypatch)\
                    .original_nvmlDeviceGetMemoryInfo(handle)\
                    .thenReturn({
                        0: mock_memory_t(total=12883853312, used=8099*MB),
                        1: mock_memory_t(total=12781551616, used=9099*MB),
                        2: mock_memory_t(total=12781551616, used=99*MB),
                    }[handle.index])

        else:  # old drivers < 510.39
            for handle in mock_gpu_handles:
                # when pynvml>=11.510, v2 API can be called but can't be used
                when(N, strict=False)\
                    .nvmlDeviceGetMemoryInfo(handle, version=ANY())\
                    .thenRaise(N.NVMLError_FunctionNotFound)
                # The v1 API will give a correct result for the v1 API
                when(N).nvmlDeviceGetMemoryInfo(handle)\
                    .thenReturn({
                        0: mock_memory_t(total=12883853312, used=8000*MB),
                        1: mock_memory_t(total=12781551616, used=9000*MB),
                        2: mock_memory_t(total=12781551616, used=0),
                    }[handle.index])

    def __getattr__(self, k):
        return self.feat[k]

    @property
    def __name__(self):
        return self.name

    def __repr__(self):
        return self.__name__


NvidiaDriverMock.INSTANCES = [
    NvidiaDriverMock('430.xx.xx',
                     nvmlDeviceGetComputeRunningProcesses_v=1,
                     nvmlDeviceGetMemoryInfo_v=1,
                     ),
    NvidiaDriverMock('450.66',
                     nvmlDeviceGetComputeRunningProcesses_v=2,
                     nvmlDeviceGetMemoryInfo_v=1,
                     ),
    NvidiaDriverMock('510.39.01',
                     nvmlDeviceGetComputeRunningProcesses_v=3,
                     nvmlDeviceGetMemoryInfo_v=2,
                     ),
]


# -----------------------------------------------------------------------------


class TestGPUStat(object):
    """A pytest class suite for gpustat."""

    def setup_method(self):
        print("")
        self.maxDiff = 4096

    def teardown_method(self):
        unstub()

    @staticmethod
    def capture_output(*args):
        f = StringIO()
        import contextlib

        with contextlib.redirect_stdout(f):  # requires python 3.4+
            try:
                gpustat.main(*args)
            except SystemExit as e:
                if e.code != 0:
                    raise AssertionError(
                        "Argparse failed (see above error message)")
        return f.getvalue()

    # -----------------------------------------------------------------------

    @pytest.mark.parametrize("nvidia_driver_version",
                             NvidiaDriverMock.INSTANCES, indirect=True)
    def test_new_query_mocked_basic(self, scenario_basic, nvidia_driver_version):
        """A basic functionality test, in a case where everything is normal."""

        gpustats = gpustat.new_query()
        fp = StringIO()
        gpustats.print_formatted(
            fp=fp, no_color=False, show_user=True,
            show_cmd=True, show_full_cmd=True, show_pid=True,
            show_fan_speed=True, show_codec="enc,dec", show_power=True,
        )

        result = fp.getvalue()
        print(result)

        unescaped = remove_ansi_codes(result)
        # remove first line (header)
        unescaped = os.linesep.join(unescaped.splitlines()[1:])

        assert unescaped == MOCK_EXPECTED_OUTPUT_FULL_PROCESS

        # verify gpustat results (not exhaustive yet)
        assert gpustats.driver_version == nvidia_driver_version.name
        g: gpustat.GPUStat = gpustats.gpus[0]
        assert g.memory_used == 8000
        assert g.power_draw == 125
        assert g.utilization == 76
        assert g.processes and g.processes[0]['pid'] == 48448

    def test_new_query_mocked_nonexistent_pid(self, scenario_nonexistent_pid):
        """
        Test a case where nvidia query returns non-existent pids (see #16, #18)
        for GPU index 2.
        """
        fp = StringIO()

        gpustats = gpustat.new_query()
        gpustats.print_formatted(fp=fp)

        ret = fp.getvalue()
        print(ret)

        # gpu 2: should ignore process id
        line = remove_ansi_codes(ret).split('\n')[3]
        assert '[2] GeForce RTX 2' in line, str(line)
        assert '99999' not in line
        assert '(Not Supported)' not in line

    @pytest.mark.parametrize("scenario_failing_one_gpu", [
        pynvml.NVMLError_GpuIsLost,
        pynvml.NVMLError_Unknown,
    ], indirect=True)
    def test_new_query_mocked_failing_one_gpu(self, scenario_failing_one_gpu):
        """Test a case where one GPU is failing (see #125)."""
        fp = StringIO()
        gpustats = gpustat.new_query()
        gpustats.print_formatted(fp=fp, show_header=False)
        ret = fp.getvalue()
        print(ret)

        lines = remove_ansi_codes(ret).split('\n')
        message = scenario_failing_one_gpu['expected_message']

        # gpu 2: failing due to unknown error
        line = lines[2]
        assert '[2] ((' + message + '))' in line, str(line)
        assert '99999' not in line
        assert '?°C,   ? %' in line, str(line)
        assert '? /     ? MB' in line, str(line)

        # other gpus should be displayed normally
        assert '[0] GeForce GTX TITAN 0' in lines[0]
        assert '[1] GeForce GTX TITAN 1' in lines[1]

    def test_attributes_and_items(self, scenario_basic):
        """Test whether each property of `GPUStat` instance is well-defined."""

        g = gpustat.new_query()[1]  # includes N/A
        print("(keys) : %s" % str(g.keys()))
        print(g)

        assert g['name'] == g.entry['name']
        assert g['uuid'] == g.uuid

        with pytest.raises(KeyError):
            g['unknown_key']

        print("uuid : %s" % g.uuid)
        print("name : %s" % g.name)
        print("memory : used %d total %d avail %d" % (
            g.memory_used, g.memory_total, g.memory_available))
        print("temperature : %d" % (g.temperature))
        print("utilization : %s" % (g.utilization))
        print("utilization_enc : %s" % (g.utilization_enc))
        print("utilization_dec : %s" % (g.utilization_dec))

    def test_main(self, scenario_basic):
        """Test whether gpustat.main() works well.
        The behavior is mocked exactly as in test_new_query_mocked().
        """
        sys.argv = ['gpustat']
        gpustat.main()

    def test_args_commandline(self, scenario_basic):
        """Tests the end gpustat CLI."""
        capture_output = self.capture_output

        def _remove_ansi_codes_and_header_line(s):
            unescaped = remove_ansi_codes(s)
            # remove first line (header)
            unescaped = os.linesep.join(unescaped.splitlines()[1:])
            return unescaped

        s = capture_output('gpustat', )
        assert _remove_ansi_codes_and_header_line(s) == MOCK_EXPECTED_OUTPUT_DEFAULT

        s = capture_output('gpustat', '--version')
        assert s.startswith('gpustat ')
        print(s)

        s = capture_output('gpustat', '--no-header')
        assert "[0]" in s.splitlines()[0]

        s = capture_output('gpustat', '-a')  # --show-all
        assert _remove_ansi_codes_and_header_line(s) == MOCK_EXPECTED_OUTPUT_FULL

        s = capture_output('gpustat', '--color')
        assert '\x0f' not in s, "Extra \\x0f found (see issue #32)"
        assert _remove_ansi_codes_and_header_line(s) == MOCK_EXPECTED_OUTPUT_DEFAULT

        s = capture_output('gpustat', '--no-color')
        unescaped = remove_ansi_codes(s)
        assert s == unescaped   # should have no ansi code
        assert _remove_ansi_codes_and_header_line(s) == MOCK_EXPECTED_OUTPUT_DEFAULT

        s = capture_output('gpustat', '--no-processes')
        assert _remove_ansi_codes_and_header_line(s) == MOCK_EXPECTED_OUTPUT_NO_PROCESSES

        s = capture_output('gpustat', '--id', '1,2')
        assert _remove_ansi_codes_and_header_line(s) == \
            os.linesep.join(MOCK_EXPECTED_OUTPUT_DEFAULT.splitlines()[1:3])

    def test_args_commandline_width(self, scenario_basic):
        capture_output = self.capture_output

        # see MOCK_EXPECTED_OUTPUT_DEFAULT
        assert len("GeForce GTX TITAN 0") == 19

        s = capture_output('gpustat', '--gpuname-width', '25')
        print("- Should have width=25")
        print(s)
        assert 'GeForce GTX TITAN 0       |' in remove_ansi_codes(s)
        assert 'GeForce RTX 2             |' in remove_ansi_codes(s)
        #                         ^012345
        #                        19

        # See #47 (since v1.0)
        print("- Should have width=10 (with truncation)")
        s = capture_output('gpustat', '--gpuname-width', '10')
        print(s)
        assert '…X TITAN 0 |' in remove_ansi_codes(s)
        assert '…rce RTX 2 |' in remove_ansi_codes(s)
        #       1234567890

        print("- Should have width=1 (too short)")
        s = capture_output('gpustat', '--gpuname-width', '1')
        print(s)
        assert '… |' in remove_ansi_codes(s)

        print("- Should have width=0: no name displayed.")
        s = capture_output('gpustat', '--gpuname-width', '0')
        print(s)
        assert '[0]  80°C' in remove_ansi_codes(s)

        print("- Invalid inputs")
        with pytest.raises(AssertionError, match="Argparse failed"):
            s = capture_output('gpustat', '--gpuname-width', '-1')
        with pytest.raises(AssertionError, match="Argparse failed"):
            s = capture_output('gpustat', '--gpuname-width', 'None')

    def test_args_commandline_showoptions(self, scenario_basic):
        """Tests gpustat CLI with a variety of --show-xxx options. """

        capture_output = self.capture_output
        print('')

        TEST_OPTS = []
        TEST_OPTS += ['-a', '-c', '-u', '-p', '-e', '-P', '-f']
        TEST_OPTS += [('-e', ''), ('-P', '')]
        TEST_OPTS += [('-e', 'enc,dec'), '-Plimit,draw']
        TEST_OPTS += ['-cup', '-cpu', '-cufP']  # 'cpuePf'

        for opt in TEST_OPTS:
            if isinstance(opt, str):
                opt = [opt]

            print('\x1b[30m\x1b[43m',  # black_on_yellow
                  '$ gpustat ' + ' '.join(shlex.quote(o) for o in opt),
                  '\x1b(B\x1b[m', sep='')
            s = capture_output('gpustat', *opt)

            # TODO: Validate output without hardcoding expected outputs
            print(s)

        # Finally, unknown args
        with pytest.raises(AssertionError):
            capture_output('gpustat', '--unrecognized-args-in-test')

    @pytest.mark.skipif(sys.platform == 'win32', reason="Do not run on Windows")
    def test_no_TERM(self, scenario_basic, monkeypatch):
        """--color should work well even when executed without TERM,
        e.g. ssh localhost gpustat --color"""
        monkeypatch.setenv("TERM", "")

        s = self.capture_output('gpustat', '--color', '--no-header').rstrip()
        print(s)
        assert remove_ansi_codes(s) == MOCK_EXPECTED_OUTPUT_DEFAULT, \
            "wrong gpustat output"

        assert '\x1b[36m' in s, "should contain cyan color code"
        assert '\x0f' not in s, "Extra \\x0f found (see issue #32)"

    def test_json_mocked(self, scenario_basic):
        gpustats = gpustat.new_query()

        fp = StringIO()
        gpustats.print_json(fp=fp)

        import json
        j = json.loads(fp.getvalue())

        from pprint import pprint
        pprint(j)

        assert j['driver_version'] == '415.27.mock'
        assert j['hostname']
        assert j['gpus']


if __name__ == '__main__':
    pytest.main()



import sys
import pytest

from gpustat import util


def test_safecall():

    def _success():
        return 42

    def _error():
        raise FileNotFoundError("oops")

    assert util.safecall(_success, error_value=None) == 42
    assert util.safecall(_error, error_value=-1) == -1

    with pytest.raises(FileNotFoundError):
        # not catched because exc_types does not match
        assert util.safecall(_error, exc_types=ValueError, error_value=-1)

    assert util.safecall(_error, error_value=-1,
                         exc_types=FileNotFoundError) == -1
    assert util.safecall(_error, error_value=-1,
                         exc_types=(FileNotFoundError, OSError)) == -1


if __name__ == '__main__':
    sys.exit(pytest.main(["-s", "-v"] + sys.argv))
